import type {SidebarsConfig} from "@docusaurus/plugin-content-docs";

const sidebars: SidebarsConfig = {
  tutorialSidebar: [{type: "autogenerated", dirName: "."}],
  apiSidebar: [{type: "autogenerated", dirName: "."}],
};

export default sidebars;
